@@ -3,14 +3,26 @@ use strict;
 
 # $Id: DBI.pm,v 1.12 2004/02/18 00:18:50 ask Exp $
 
-BEGIN { eval { require Apache } }
-BEGIN { eval { require mod_perl2; require Apache2::Module; } }
+use constant MP2 => (exists $ENV{MOD_PERL_API_VERSION} && 
+    $ENV{MOD_PERL_API_VERSION} == 2) ? 1 : 0;
+
+BEGIN { 
+    if (MP2) {
+        require mod_perl2; 
+        require Apache2::Module; 
+        require Apache2::ServerUtil;
+    }
+    elsif (defined $modperl::VERSION && $modperl::VERSION > 1 && 
+        $modperl::VERSION < 1.99) {
+        require Apache;
+    }
+}
 use DBI ();
 use Carp qw(carp);
 
 require_version DBI 1.00;
 
-$Apache::DBI::VERSION = '0.99';
+$Apache::DBI::VERSION = '0.9901';
 
 # 1: report about new connect
 # 2: full debug output
@@ -22,8 +34,9 @@ my @ChildConnect; # connections to be established when a new httpd child is crea
 my %Rollback;     # keeps track of pushed PerlCleanupHandler which can do a rollback after the request has finished
 my %PingTimeOut;  # stores the timeout values per data_source, a negative value de-activates ping, default = 0
 my %LastPingTime; # keeps track of last ping per data_source
-my $Idx;          # key of %Connected and %Rollback.
 
+# Check to see if we need to reset TaintIn and TaintOut
+my $TaintInOut = ($DBI::VERSION>=1.31)?1:0;
 
 # supposed to be called in a startup script.
 # stores the data_source of all connections, which are supposed to be created upon
@@ -32,19 +45,19 @@ my $Idx;          # key of %Connected and %Rollback.
 sub connect_on_init { 
     # provide a handler which creates all connections during server startup
 
-		if ($ENV{MOD_PERL_API_VERSION} == 2) {
-			if (!@ChildConnect) {
-				require Apache2::ServerUtil;
-				my $s = Apache2::ServerUtil->server;
-				$s->push_handlers(PerlChildInitHandler => \&childinit);
-			}
-		}
-		else {
-   		carp "Apache.pm was not loaded\n" and return unless $INC{'Apache.pm'};
-   		if (!@ChildConnect and Apache->can('push_handlers')) {
-     	  Apache->push_handlers(PerlChildInitHandler => \&childinit);
-   		}
-		}
+    if (MP2) {
+        if (!@ChildConnect) {
+             my $s = Apache2::ServerUtil->server;
+             $s->push_handlers(PerlChildInitHandler => \&childinit);
+        }
+    }
+    else {
+        carp "Apache.pm was not loaded\n" and return unless $INC{'Apache.pm'};
+        if (!@ChildConnect and Apache->can('push_handlers')) {
+            Apache->push_handlers(PerlChildInitHandler => \&childinit);
+        }
+    }
+
     # store connections
     push @ChildConnect, [@_];
 }
@@ -76,7 +89,7 @@ sub connect {
     my $dsn    = "dbi:$drh->{Name}:$args[0]";
     my $prefix = "$$ Apache::DBI            ";
 
-    $Idx = join $;, $args[0], $args[1], $args[2];
+    my $Idx = join $;, $args[0], $args[1], $args[2]; # key of %Connected and %Rollback.
 
     # the hash-reference differs between calls even in the same
     # process, so de-reference the hash-reference 
@@ -97,7 +110,7 @@ sub connect {
     # unpredictable query results.
 
     ## See: http://perl.apache.org/docs/2.0/user/porting/compat.html#C__Apache__Server__Starting__and_C__Apache__Server__ReStarting_
-    if (exists $ENV{MOD_PERL_API_VERSION} && $ENV{MOD_PERL_API_VERSION} == 2) {
+    if (MP2) {
         require Apache2::ServerUtil;
         if (Apache2::ServerUtil::restart_count() == 1) {
           print STDERR "$prefix skipping connection during server startup, read the docu !!\n" if $Apache::DBI::DEBUG > 1;
@@ -110,19 +123,20 @@ sub connect {
         }   
     }
 
-    # this PerlCleanupHandler is supposed to initiate a rollback after the script has finished if AutoCommit is off.
-    my $needCleanup = ($Idx =~ /AutoCommit[^\d]+0/) ? 1 : 0;
-    if(!$Rollback{$Idx} and $needCleanup and Apache->can('push_handlers')) {
+    # this PerlCleanupHandler is supposed to initiate a rollback after the
+    # script has finished if AutoCommit is off.  however, cleanup can only
+    # be determined at end of handle life as begin_work may have been called
+    # to temporarily turn off AutoCommit.
+    if(!$Rollback{$Idx} and Apache->can('push_handlers')) {
         print STDERR "$prefix push PerlCleanupHandler \n" if $Apache::DBI::DEBUG > 1;
-				if ($ENV{MOD_PERL_API_VERSION} == 2) {
-					require Apache2::ServerUtil;
-					my $s = Apache2::ServerUtil->server;
-					$s->push_handlers("PerlCleanupHandler", \&cleanup);
-				}
-				else {
-        	Apache->push_handlers("PerlCleanupHandler", \&cleanup);
-				}
-        # make sure, that the rollback is called only once for every 
+        if (MP2) {
+            my $s = Apache2::ServerUtil->server;
+            $s->push_handlers("PerlCleanupHandler", sub { cleanup($Idx) });
+        }
+        else {
+            Apache->push_handlers("PerlCleanupHandler", sub { cleanup($Idx) });
+        }
+        # make sure, that the rollback is called only once for every
         # request, even if the script calls connect more than once
         $Rollback{$Idx} = 1;
     }
@@ -131,9 +145,11 @@ sub connect {
     $PingTimeOut{$dsn}  = 0 unless $PingTimeOut{$dsn};
     $LastPingTime{$dsn} = 0 unless $LastPingTime{$dsn};
     my $now = time;
-    my $needping = (($PingTimeOut{$dsn} == 0 or $PingTimeOut{$dsn} > 0)
-		    and (($now - $LastPingTime{$dsn}) >= $PingTimeOut{$dsn})
-		   ) ? 1 : 0;
+    # Must ping if TimeOut = 0 else base on time
+    my $needping = ($PingTimeOut{$dsn} == 0 or
+                    ($PingTimeOut{$dsn} > 0 and
+                     $now - $LastPingTime{$dsn} > $PingTimeOut{$dsn})
+                   ) ? 1 : 0;
     print STDERR "$prefix need ping: ", $needping == 1 ? "yes" : "no", "\n" if $Apache::DBI::DEBUG > 1;
     $LastPingTime{$dsn} = $now;
 
@@ -144,6 +160,10 @@ sub connect {
     # RaiseError being on and the handle is invalid.
     if ($Connected{$Idx} and (!$needping or eval{$Connected{$Idx}->ping})) {
         print STDERR "$prefix already connected to '$Idx'\n" if $Apache::DBI::DEBUG > 1;
+
+        # Force clean up of handle in case previous transaction failed to clean up the handle
+        &reset_startup_state($Idx);
+
         return (bless $Connected{$Idx}, 'Apache::DBI::db');
     }
 
@@ -153,6 +173,9 @@ sub connect {
     $Connected{$Idx} = $drh->connect(@args);
     return undef if !$Connected{$Idx};
 
+    # store the parameters of the initial connection in the handle
+    set_startup_state($Idx);
+
     # return the new database handle
     print STDERR "$prefix new connect to '$Idx'\n" if $Apache::DBI::DEBUG;
     return (bless $Connected{$Idx}, 'Apache::DBI::db');
@@ -180,6 +203,7 @@ sub childinit {
 # Note: the PerlCleanupHandler runs after the response has been sent to the client
 
 sub cleanup {
+    my $Idx = shift;
     my $prefix = "$$ Apache::DBI            ";
     print STDERR "$prefix PerlCleanupHandler \n" if $Apache::DBI::DEBUG > 1;
     my $dbh = $Connected{$Idx};
@@ -190,6 +214,49 @@ sub cleanup {
     1;
 }
 
+# Store the default start state of each dbh in the handle
+# Note: This uses private_Apache_DBI hash ref to store it in the handle itself
+
+sub set_startup_state {
+    my $Idx = shift;
+    foreach my $key qw{ AutoCommit Warn CompatMode InactiveDestroy 
+                        PrintError RaiseError HandleError 
+                        ShowErrorStatement TraceLevel FetchHashKeyName 
+                        ChopBlanks LongReadLen LongTruncOk 
+                        Taint Profile} {
+        $Connected{$Idx}->{private_Apache_DBI}{$key} = $Connected{$Idx}->{$key};
+    }
+    if ($TaintInOut) {
+        foreach my $key qw{ TaintIn TaintOut } {
+            $Connected{$Idx}->{private_Apache_DBI}{$key} = $Connected{$Idx}->{$key};
+        }
+    }
+    1;
+}
+
+
+# Restore the default start state of each dbh
+
+sub reset_startup_state {
+    my $Idx = shift;
+    # Rollback current transaction if currently in one
+    $Connected{$Idx}->{Active} and !$Connected{$Idx}->{AutoCommit} and eval {$Connected{$Idx}->rollback};
+
+    foreach my $key qw{ AutoCommit Warn CompatMode InactiveDestroy 
+                        PrintError RaiseError HandleError 
+                        ShowErrorStatement TraceLevel FetchHashKeyName 
+                        ChopBlanks LongReadLen LongTruncOk 
+                        Taint Profile } {
+        $Connected{$Idx}->{$key} = $Connected{$Idx}->{private_Apache_DBI}{$key};
+    }
+    if ($TaintInOut) {
+        foreach my $key qw{ TaintIn TaintOut } {
+            $Connected{$Idx}->{$key} = $Connected{$Idx}->{private_Apache_DBI}{$key};
+        }
+    }
+    1;
+}
+
 
 # This function can be called from other handlers to perform tasks on all cached database handles.
 
@@ -218,7 +285,7 @@ sub all_handlers {
 
 
 # prepare menu item for Apache::Status
-if (exists $ENV{MOD_PERL_API_VERSION} && $ENV{MOD_PERL_API_VERSION} == 2) {
+if (MP2) {
 	Apache2::Status->menu_item(
     'DBI' => 'DBI connections',
     sub {
@@ -345,8 +412,10 @@ every script. In order to avoid inconsistencies in the database in case
 AutoCommit is off and the script finishes without an explicit rollback, the 
 Apache::DBI module uses a PerlCleanupHandler to issue a rollback at the
 end of every request. Note, that this CleanupHandler will only be used, if 
-the initial data_source sets AutoCommit = 0. It will not be used, if AutoCommit 
-will be turned off, after the connect has been done. 
+the initial data_source sets AutoCommit = 0 or AutoCommit is turned off, after 
+the connect has been done (ie begin_work). However, because a connection may have
+set other parameters, the handle is reset to its initial connection state before 
+it is returned for a second time.
 
 This module plugs in a menu item for Apache::Status or Apache2::Status. 
 The menu lists the current database connections. It should be considered 
